//(c) 2016 Dmitry Grinberg   http://dmitry.gr   me@dmitry.gr   see LICENSE.txt

#include "wincompat.h"
#ifdef WIN32
#include <windows.h>
#else
#include <sys/time.h>
#endif
#include "hidapi/hidapi/hidapi.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "flash.h"
#include "kln.h"
#include "../comms.h"



static uint32_t mFlashSz, mFlashErzblockSz, mKey, mMagix;
static hid_device *mHandle = NULL;




static uint64_t getMsecsSinceSometime(void)
{
#ifdef WIN32
	FILETIME ft;
	uint64_t ret;
	
	
	GetSystemTimeAsFileTime(&ft);
	ret = ft.dwHighDateTime;
	ret <<= 32;
	ret |= ft.dwLowDateTime;
	ret /= 10000;
	return ret;
	
#else
	struct timeval tv;
	if(gettimeofday(&tv, NULL))
		return 0;
	return (unsigned long)((tv.tv_sec * 1000ul) + (tv.tv_usec / 1000ul));
#endif
}

static bool flashProgrammerDoOneCmd(const struct CommsPacket *pktTx, struct CommsPacket *pktRx)
{
	unsigned char bufTx[sizeof(struct CommsPacket) + 1];
	struct CommsPacket *pktTxReal = (struct CommsPacket*)(bufTx + 1);
	int nRetries = 8,  rxLen;
	
	*pktTxReal = *pktTx;
	pktTxReal->crc = commsPacketCalc(pktTxReal);
	bufTx[0] = 0;
	
	while(nRetries--) {
		
		if (hid_write(mHandle, bufTx, 1 + sizeof(struct CommsPacket)) < 0)
			return false;
		
		while ((rxLen = hid_read_timeout(mHandle, (char*)pktRx, sizeof(struct CommsPacket), 2000)) > 0) {
			
			if (pktRx->cmd != pktTxReal->cmd)
				continue;
			
			if (rxLen != sizeof(struct CommsPacket))
				continue;
			
			if (commsPacketCalc(pktRx) != pktRx->crc)
				break;
			
			return true;
		}
	}
	
	return false;
}

static void flashProgrammerDrainOsBuf(void)
{
	//At least linux will queue up bufers from device even while we're not expecting them. This will drain them.
	struct CommsPacket pkt;
	
	while (hid_read_timeout(mHandle, (char*)&pkt, sizeof(struct CommsPacket), 50) > 0);
}

static bool flashProgrammerShowUsbInfo(void)
{
	#define USB_DEV_MAX_LEN	256
	wchar_t manuf[USB_DEV_MAX_LEN], prod[USB_DEV_MAX_LEN];
	
	if (hid_get_manufacturer_string(mHandle, manuf, USB_DEV_MAX_LEN) || hid_get_product_string(mHandle, prod, USB_DEV_MAX_LEN)) {
		fprintf(stderr, "NO USB INFO AVAIL\n");
		return false;
	}
	
	fprintf(stderr, "Dev USB info: '%ls' by '%ls'\n", prod, manuf);
	return true;
}

static bool flashGetCardInfo(void)
{
	struct CommsPacket tx, rx;
	
	tx.cmd = PACKET_TYPE_GET_INFO;
	
	if (flashProgrammerDoOneCmd(&tx, &rx)) {
		
		fprintf(stderr, "PROGRAMMER INFO:\n  Version:     %08X\n  Flash found: %s\n", rx.info.version, rx.info.compatible ? "YES" : "NO");
		if (!rx.info.compatible)
			return false;
			
		fprintf(stderr, "CARD INFO:\n");
		fprintf(stderr, "  Key:         %08X\n  Magic:       %08X\n", rx.info.key, rx.info.magicval);
		fprintf(stderr, "  Size:        %08X\n  Erz Blk Sz:  %08X\n", rx.info.flashSz, rx.info.flashErzBlkSz);
		mFlashSz = rx.info.flashSz;
		mFlashErzblockSz = rx.info.flashErzBlkSz;
		mKey = rx.info.key;
		mMagix = rx.info.magicval;
		
		return true;
	}
	return false;
}

static bool flashRead(uint32_t addr, uint8_t *dst)
{
	struct CommsPacket tx, rx;
	
	tx.cmd = PACKET_TYPE_DO_READ;
	tx.data.addr = addr;
	
	if (flashProgrammerDoOneCmd(&tx, &rx)) {
		
		if (rx.data.addr == addr) {	//success
			
			memcpy(dst, rx.data.bytes, COMMS_DATA_TX_SZ);
			return true;
		}
	}
	return false;
}

static bool flashVppControl(bool on)
{
	struct CommsPacket tx, rx;
	
	tx.cmd = PACKET_TYPE_VPP_CTRL;
	tx.on = on;
	
	if (flashProgrammerDoOneCmd(&tx, &rx)) {
		
		return !rx.on == !on;
	}
	return false;
}

static bool flashErase(uint32_t addr)
{
	struct CommsPacket tx, rx;
	
	tx.cmd = PACKET_TYPE_DO_ERASE;
	tx.erase.addr = addr;
	
	if (flashProgrammerDoOneCmd(&tx, &rx)) {
		
		return rx.success;
	}
	return false;
}

static bool flashWrite(uint32_t addr, const uint8_t *data)
{
	struct CommsPacket tx, rx;
	
	tx.cmd = PACKET_TYPE_DO_WRITE;
	tx.data.addr = addr;
	memcpy(tx.data.bytes, data, COMMS_DATA_TX_SZ);
	
	if (flashProgrammerDoOneCmd(&tx, &rx)) {
		
		return rx.success;
	}
	return false;
}

static bool flashInit(void)
{
	int ret;
	
	if (mHandle)				//only init once
		return true;
	
	//init lib
	ret = hid_init();
	if (ret) {
		fprintf(stderr, "Failed to load hid lib\n");
		goto out;
	}

	//Find the device
	mHandle = hid_open(0x4744, 0x5567, NULL);
	if (!mHandle) {
		fprintf(stderr, "Failed to find device\n");
		goto out_with_hidlib;
	}

	//drain buffer just in case
	flashProgrammerDrainOsBuf();
	
	//get USB device info
	if (!flashProgrammerShowUsbInfo())
		goto out_with_handle;
	
	//get memory card info
	if (!flashGetCardInfo())
		goto out_with_handle;

	return true;
	
out_with_handle:
	hid_close(mHandle);
	mHandle = NULL;
	
out_with_hidlib:
	hid_exit();

out:
	return false;
}

static void flashShowProgress(uint32_t addr, uint32_t total, uint32_t doneInTimeSlice, uint64_t timesliceLen)
{
	if (addr < total)
		fprintf(stderr, "0x%08lX/0x%08lX (%3lu%%)", (unsigned long)addr, (unsigned long)total, (unsigned long)((uint64_t)addr * 100 / total));
	else
		fprintf(stderr, "DONE       0x%08lX (100%%)", (unsigned long)total);
	if (timesliceLen)
		fprintf(stderr, " %u b/s     ", (unsigned)((uint64_t)(doneInTimeSlice * 1000 + timesliceLen / 2) / timesliceLen));
	else
		fprintf(stderr, "                       ");
	fprintf(stderr, "\r");
}

bool flashMemoryRead(uint8_t *dst, uint32_t *dstSzP)		//*dstSzP on input is mas size we can produce produce!
{
	uint64_t totalTime, time;
	uint32_t i, j;
	char lt, ct = 0;
	
	if (!flashInit()) {
		fprintf(stderr, "Failed to establish comms with updater\n");
		return false;
	}
	
	if (mFlashSz > *dstSzP) {
		
		fprintf(stderr, "Cannot fit %lu bytes of flash into %lu bytes of buffer!\n", (unsigned long)mFlashSz, (unsigned long)*dstSzP);
		return false;
	}
	
	*dstSzP = mFlashSz;

	fprintf(stderr, "Reading...\n");
	totalTime = getMsecsSinceSometime();
	time = 0;
	for (i = 0; i < mFlashSz; i += COMMS_DATA_TX_SZ) {
		uint8_t buf[COMMS_DATA_TX_SZ];
		
		lt = (char)((getMsecsSinceSometime() - totalTime) >> 10) & 1;
		if (lt != ct)
			flashShowProgress(i, mFlashSz, COMMS_DATA_TX_SZ, time);
		ct = lt;
		
		time = getMsecsSinceSometime();
		if (!flashRead(i, buf)) {
			fprintf(stderr, "\nERROR\n");
			return false;
		}
		time = getMsecsSinceSometime() - time;
		
		j = mFlashSz - i;
		if (j > COMMS_DATA_TX_SZ)
			j = COMMS_DATA_TX_SZ;
		
		memcpy(dst + i, buf, j);
	}
	totalTime = getMsecsSinceSometime() - totalTime;
	flashShowProgress(mFlashSz, mFlashSz, mFlashSz, totalTime);
	fprintf(stderr, "\n\n");
	
	return true;
}

bool flashMemoryWrite(uint8_t *data, uint32_t writeLen, bool allowDiffKeys, bool useCardKeys)
{
	struct KlnDecryptedUpdateHeader *flashable = (struct KlnDecryptedUpdateHeader*)data, backup;
	uint64_t totalTime, time;
	bool ret = false;
	char lt, ct = 0;
	uint32_t i, j;

	if (!flashInit()) {
		fprintf(stderr, "Failed to establish comms with updater\n");
		return false;
	}

	if (writeLen > mFlashSz) {
		fprintf(stderr, "Cannot fit %lu bytes of buffer into %lu bytes of flash!\n", (unsigned long)writeLen, (unsigned long)mFlashSz);
		return false;
	}
	
	if (writeLen < sizeof(struct KlnDecryptedUpdateHeader)) {
		fprintf(stderr, "The %lu bytes of buffer cannot be an update (too small)!\n", (unsigned long)writeLen);
		return false;
	}
	
	if (!flashVppControl(true)) {
		fprintf(stderr, "Failed to enable Vpp\n");
		return false;
	}
	
	backup = *flashable;
	if (flashable->key == 0xfffffffful && flashable->magix == 0xfffffffful && mKey != 0xfffffffful && mMagix != 0xfffffffful) {
		fprintf(stderr, "Writing current card's key and magix (%08lX %08lX) into update\n", (unsigned long)mKey, (unsigned long)mMagix);
		flashable->key = mKey;
		flashable->magix = mMagix;
	}
	
	if (flashable->key != mKey || flashable->magix != mMagix) {
		fprintf(stderr,	"Update has key %08lX and magix %lX, card has %lX and %lX!\n",
						(unsigned long)flashable->key, (unsigned long)flashable->magix, (unsigned long)mKey, (unsigned long)mMagix);
		
		if (mKey == 0xfffffffful && mMagix == 0xfffffffful) {
			
			fprintf(stderr, "Forcing use of update's key as card's key is invalid\n");
		}
		else if (!allowDiffKeys) {
			fprintf(stderr, " Refusing to perform write. If you're sure you want to do this, pass in\n"
							"the \"-Xc\" or the \"-Xu\" flag!");
			goto out;
		}
		else {
			fprintf(stderr, " Warning overridden! Using %s's key & magix.\n", useCardKeys ? "card" : "update");
			
			if (useCardKeys) {
				flashable->key = mKey;
				flashable->magix = mMagix;
			}
		}
	}

	if (flashable->key == 0xfffffffful && flashable->magix == 0xfffffffful) {
		fprintf(stderr, "No valid key and magix found. Using failsafes!\n");
		flashable->key = 0x6d59ebcful;
		flashable->magix = 0x46eae0dbul;
	}

	fprintf(stderr, "Writing update with key %08lX and magix %08lX\n",
					(unsigned long)flashable->key, (unsigned long)flashable->magix);

	fprintf(stderr, "Erasing...\n");
	totalTime = getMsecsSinceSometime();
	time = 0;
	for (i = 0; i < mFlashSz; i += mFlashErzblockSz) {
		
		flashShowProgress(i, mFlashSz, mFlashErzblockSz, time);
		
		time = getMsecsSinceSometime();
		if (!flashErase(i)) {
			fprintf(stderr, "\nERROR erasing\n");
			goto out;
		}
		time = getMsecsSinceSometime() - time;
	}
	totalTime = getMsecsSinceSometime() - totalTime;
	flashShowProgress(mFlashSz, mFlashSz, mFlashSz, totalTime);
	fprintf(stderr, "\n\n");

	fprintf(stderr, "Writing...\n");
	totalTime = getMsecsSinceSometime();
	time = 0;
	for (i = 0; i < writeLen; i += COMMS_DATA_TX_SZ) {
		
		uint8_t buf[COMMS_DATA_TX_SZ];
		
		j = writeLen - i;
		if (j >= COMMS_DATA_TX_SZ)
			j = COMMS_DATA_TX_SZ;
		else
			memset(buf, 0xff, sizeof(buf));
		
		memcpy(buf, data + i, j);
		
		lt = (char)((getMsecsSinceSometime() - totalTime) >> 10) & 1;
		if (lt != ct)
			flashShowProgress(i, mFlashSz, COMMS_DATA_TX_SZ, time);
		ct = lt;
		
		
		time = getMsecsSinceSometime();
		if (!flashWrite(i, buf)) {
			fprintf(stderr, "\nERROR writing\n");
			goto out;
		}
		time = getMsecsSinceSometime() - time;
	}
	totalTime = getMsecsSinceSometime() - totalTime;
	flashShowProgress(writeLen, writeLen, writeLen, totalTime);
	fprintf(stderr, "\n\n");
	
	ret = true;
	
	//if we wrote the update, update our idea of car's current "key" and "magix"
	mKey = flashable->key;
	mMagix = flashable->magix;
	
out:
	
	*flashable = backup;
	(void)flashVppControl(false);	//just in case
	return ret;
}


void flashDeinit(void)
{
	if (mHandle) {
		(void)flashVppControl(false);	//just in case
		hid_close(mHandle);
		hid_exit();
	}
	mHandle = NULL;
}