//(c) 2016 Dmitry Grinberg   http://dmitry.gr   me@dmitry.gr   see LICENSE.txt

#include <avr/io.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include "usb.h"
#include "hw.h"

//TODO: test mode?
//TODO: callback from INT to app to handle other EP0 requests? (HID can use it)


#define GET_STATUS			0
#define CLEAR_FEATURE			1
#define SET_FEATURE			3
#define SET_ADDRESS			5
#define GET_DESCRIPTOR			6
#define SET_DESCRIPTOR			7	//todo
#define GET_CONFIGURATION		8
#define SET_CONFIGURATION		9
#define GET_INTERFACE			10	//todo
#define SET_INTERFACE			11	//todo
#define SYNCH_FRAME			12	//todo

#define BM_REQ_IS_TO_HOST(r)		(!((r) & 0x80))


static struct{
	uint8_t configured	: 1;	//set if we'v ebeen configured
	uint8_t selfPowered	: 1;	//set if we're self powered
	uint8_t remoteWakeup	: 1;	//set if we've been told to enable remote wakeup
	
	uint8_t gConfigNum;		//whicih config is currently in use
	
	volatile uint32_t ticks;
}gCfg;


static void usbPrvConfigEp0(void){
	
	usbEpConfigure(0, EP_TYPE_CONTROL | EP_SIZE_64 | EP_ENABLED);
	UEIENX = (1 << RXSTPE);	//int on RX, this relies on UENUM staying at zero after usbEpConfigure()
}

void usbInit(bool lowSpeed, bool selfPowered){

	USBCON = (1 << USBE);			//USB on
	USBCON = (1 << USBE) | (1 << OTGPADE);	//USB on
	UDADDR = 0;
	UDCON = lowSpeed ? 0b00000101 : 0b00000001;
	gCfg.configured = 0;
	gCfg.selfPowered = selfPowered;
	gCfg.gConfigNum = 0;
	
	usbPrvConfigEp0();
	UDIEN = (1 << EORSTE) | (1 << SOFE);			//int on reset end
}

void usbRemoteWake(){

	UDCON |= 0b00000010;
	while(UDCON & 0b00000010);
}

void usbAttachControl(bool attach){

	if(attach) UDCON &= (uint8_t)~0b00000001;
	else UDCON |= 0b00000001; 	
}

bool usbEpConfigure(uint8_t ep, uint8_t cfg){
	
	bool ret = true;
	
	UENUM = ep;
	if(cfg){
		UECONX = 0b00001001;
		
		UECFG0X = cfg & 0b11000001;
		UECFG1X = (cfg & 0b00111010) << 1;	//set config
		
		UECFG1X = ((cfg & 0b00111010) << 1) | 0b00000010;			//alloc memory
		if(UESTA0X & 0b10000000) return true;
		ret = false;
	}
	
	UECONX = 0b00000000;
	UECFG1X = 0b00000000;
	return ret;
}

void usbEpStall(uint8_t ep, bool stall){

	UENUM = ep;
	UECONX = (UECONX & 0b00000001) | (stall ? 0b00100000 : 0b00010000);
}

bool usbEpStalled(uint8_t ep){
	
	UENUM = ep;
	return !!(UECONX & 0b00100000);
}

uint8_t usbEpGetNumUsedBanks(uint8_t ep){

	UENUM = ep;
	
	return UESTA0X & 3;
}

bool usbEpCanSend(uint8_t ep){

	UENUM = ep;
	return !!(UEINTX & (1 << RWAL));
}

static inline void usbPrvTxRdyWait(void){
	
	while (!(UEINTX & (1 << TXINI)));
}

static inline void usbPrvTxDo(void){
	
	UEINTX = (uint8_t)~(1 << TXINI);
}

void usbEpDoSendNow(uint8_t ep){
	
	UENUM = ep;
	UEINTX = (1 << RWAL) | (1 << NAKOUTI) | (1 << RXSTPI) | (1 << STALLEDI);
}

bool usbEpHasData(uint8_t ep){
	
	UENUM = ep;
	return (UEINTX >> RXOUTI) & 1;
}

uint16_t usbEpNumBytesAvail(uint8_t ep){			//only use if above said there is data

	UENUM = ep;
	return UEBCX;
}

void usbEpRxBytes(uint8_t ep, uint8_t *buf, uint16_t num){	//read more than there is avail at your own peril
	UENUM = ep;
	while(num--){
		
		if(buf) *buf++ = UEDATX;
		else (void)UEDATX;
	}
	UEINTX = (uint8_t)~((1 << FIFOCON) | (1 << RXOUTI)); 	//clear these flags
}

void usbEpDoSend(uint8_t ep, const void* dataP, uint16_t len){
	
	const uint8_t* data = dataP;
	uint16_t epSz, now;
	bool needZLP;
	
	UENUM = ep;
	epSz = 8 << ((UECFG1X >> 4) & 7);
	do{
		needZLP = false;
		
		now = len;
		if(now > epSz) now = epSz;
		len -= now;
		
		if(!len && now == epSz) needZLP = true;
		
		if(ep) while(!(UEINTX & (1 << RWAL)));	//wait for a slot in the TX buffer (ep != 0)
		else usbPrvTxRdyWait();			//wait for TX ready (ep == 0)
		
		while(now--) UEDATX = *data++;
		
		if(ep) UEINTX = (1 << RWAL) | (1 << NAKOUTI) | (1 << RXSTPI) | (1 << STALLEDI);	//send packet (think: it's not the bits I set that matter, it's the ones I clear)
		else usbPrvTxDo();
		
	}while(len || needZLP);
}

static void usbPrvSendDescr(uint8_t descrTyp, uint8_t descrIdx, uint16_t langID, uint16_t len){

	uint16_t ofst = 0;
	uint8_t t, i;
	
	do{
		t = (len < USB_EP0_SIZE)? len : USB_EP0_SIZE;
		len -= t;
		i = t;
		
		while (!((t = UEINTX) & ((1 << TXINI) | (1 << RXOUTI))));	//wait for host to be ready for an  IN packet
		if (t & (1 << RXOUTI)) return;					//something went very wrong - we got a packet while we didn't expect it
	
		while(i--) UEDATX = usbCbkGetDescriptor(descrTyp, descrIdx, langID, ofst++);
		usbPrvTxDo();
		
	}while(len || t == USB_EP0_SIZE);	//second condition assures a ZLP if we were exactly the right size for a packet boundary
}

bool usbEp0canSend(void)
{
	UENUM = 0;
	return !!(UEINTX & (1 << TXINI));
}

ISR(USB_COM_vect){	//handle EP0 only, since it is entirely managed by the driver. Let everyone else deal with PIO :)
			//TODO: maybe re-enable interrupts here this can run a while...not cool

	uint8_t bmRequestType, bRequest, wValueL, wValueH, wIndexL, wIndexH, wLengthL, wLengthH;
	uint8_t sta, t = 0, olduenum;
	uint16_t L;
	
	olduenum = UENUM;	//if we interrupt code that was working with a different endpoint, we shoudl be sure to leave it still with the same endpoint, lest it mess up EP0 for us
	UENUM = 0;
	sta = UEINTX;

	if(sta & (1 << RXSTPI)){
		
		bmRequestType = UEDATX;
		bRequest = UEDATX;
		wValueL = UEDATX;
		wValueH = UEDATX;
		wIndexL = UEDATX;
		wIndexH = UEDATX;
		wLengthL = UEDATX;
		wLengthH = UEDATX;
		UEINTX = ~((1 << RXSTPI) | (1 << RXOUTI) | (1 << TXINI)); 	//clear these flags
		
		if(BM_REQ_IS_STD(bmRequestType)) switch(bRequest){
			case GET_STATUS:
				usbPrvTxRdyWait();
				if(BM_REQ_IS_DEVIC(bmRequestType)){
					if(gCfg.selfPowered) t++;
					if(gCfg.remoteWakeup) t |= 2;
				}
				else if(BM_REQ_IS_ENDPT(bmRequestType)){
					//the only existing endpoint feature is "HALT", which is a persistent STALL - return its status
					UENUM = wIndexL & 0x0F;
					if(UECONX & (1 << STALLRQ)) t++;
					UENUM = 0;
				}
				else if(!BM_REQ_IS_INTFC(bmRequestType)) goto fail;	//invalid request
				UEDATX = t;
				UEDATX = 0;
				usbPrvTxDo();	//send the reply
				break;
			
			case CLEAR_FEATURE:
				
				t = 1;
				//fallthrough
			
			case SET_FEATURE:
				
				//now t is 0 if set, 1 to clear the feature
				if(wValueH) goto fail;					//invalid request
				if(BM_REQ_IS_DEVIC(bmRequestType) && wValueL == 1){	//remote wake up
					
					usbPrvTxDo();	//send a ZLP
					gCfg.remoteWakeup = 1 - t;
				}
				else if(BM_REQ_IS_DEVIC(bmRequestType) && wValueL == 2){	//test mode
					
					usbPrvTxDo();	//send a ZLP
					//TODO: test mode
				}
				else if(BM_REQ_IS_ENDPT(bmRequestType) && wValueL == 0){	//halt
					
					usbPrvTxDo();	//send a ZLP
					UENUM = wIndexL & 0x0F;
					
					if(t){	//clear halt
					
						//unstall endpt
						UECONX = (1 << STALLRQC) | (1 << RSTDT) | (1 << EPEN);
						UERST = (1 << UENUM);	//resetn endpt
						UERST = 0;
					}
					else{	//set halt
					
						//stall endpt
						UECONX = (1 << STALLRQ) | (1 << EPEN);
					}
					UENUM = 0;
				}
				else goto fail;	//invalid request
				break;
			
			case SET_ADDRESS:
				
				UDADDR = wValueL;
				usbPrvTxDo();		//send a ZLP
				usbPrvTxRdyWait();	//wait for it to go out
				UDADDR |= 1 << ADDEN;
				gCfg.configured = 1;
				break;
			
			case GET_DESCRIPTOR:
				
				//wValueH is descriptor type
				//wValueL is descriptor index (only for config & string descriptors, else zero)
				//wIndex is languageID for string, zero else
				//wLength is max num bytes to return
				
				L = usbCbkGetDescriptor(wValueH, wValueL, (((uint16_t)wIndexH) << 8) + wIndexL, OFST_VAL_SIZE_H);
				L <<= 8;
				L |= usbCbkGetDescriptor(wValueH, wValueL, (((uint16_t)wIndexH) << 8) + wIndexL, OFST_VAL_SIZE_L);
				
				if(!L) goto fail;
				
				if(L > (((uint16_t)wLengthH) << 8) + wLengthL) L = (((uint16_t)wLengthH) << 8) + wLengthL;
				if(L) usbPrvSendDescr(wValueH, wValueL, (((uint16_t)wIndexH) << 8) + wIndexL, L);
				break;
			
			case GET_CONFIGURATION:
				
				UEDATX = gCfg.gConfigNum;
				usbPrvTxDo();
				break;
			
			case SET_CONFIGURATION:
				
				if(bmRequestType) goto fail;
				if(wValueL && !usbCbkSetConfig(wValueL)) goto fail;
				usbPrvTxDo();		//send a ZLP
				gCfg.gConfigNum = wValueL;
				if(wValueL) usbCbkSetConfig(wValueL | CFG_NUM_OR_AFTER);
				else for(wValueL = 1; wValueL < USB_NUM_EP + 1; wValueL++) usbEpConfigure(wValueL, 0);
				break;
		}
		else if(!usbCbkCustomReqEp0(bmRequestType, bRequest, wValueL, wValueH, wIndexL, wIndexH, wLengthL, wLengthH)) goto fail;
	}
	UENUM = olduenum;
	return;
	
fail:
	//if we got here, we've not handled this message - stall
	UECONX = (1 << STALLRQ) | (1 << EPEN);
	UENUM = olduenum;
}

ISR(USB_GEN_vect){
	
	if(UDINT & (1 << EORSTI)){	//reset over?
		gCfg.configured = 0;
		usbPrvConfigEp0();
	}
	if(UDINT & (1 << SOFI)){	//SOF
		gCfg.ticks++;
	}
	UDINT = 0;
}

bool usbIsConfigured(void)
{
	return gCfg.configured;
}

uint32_t usbGetTicks(void)
{
	uint32_t read;
	
	do{
		read = gCfg.ticks;
	}while(gCfg.ticks != read);
	
	return read;
}